#include <errno.h>
#include <sys/types.h>
#include <sys/acct.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

char *fname = NULL;
int fd = -1;

char *gen_temp_fname(int f_nr)
{
	char *buffer = (char *)malloc(128);
	if (!buffer)
	{
		fprintf(stderr, "out of memory\n");
		exit(1);
	}

	sprintf(buffer, ".%d.%d._acctail", getpid(), f_nr);

	return buffer;
}

void sigh(int sig)
{
	acct(NULL);
	if (fd != -1)
		close(fd);
	if (fname)
	{
		if (unlink(fname) == -1)
			fprintf(stderr, "failed to delete file %s: %d\n", fname, errno);
	}

	exit(1);
}

unsigned long int expand_compt(comp_t val)
{
	unsigned long int exponent;
	unsigned long int mantisa;

	exponent = (val >> 13) & 07;
	mantisa = val & 017777;

	while (exponent-- > 0)
		mantisa <<= 3;

	return mantisa;
}

double compt_to_msec(comp_t timeval)
{
	return ((double)(expand_compt(timeval) * 1000)) / AHZ;
}

void usage(void)
{
	fprintf(stderr, "-s  time before the temporary file is deleted\n");
#ifndef __APPLE__
	fprintf(stderr, "-d  show details\n");
#endif
	fprintf(stderr, "-h  this help\n");
}

int main(int argc, char *argv[])
{
	int delay = 3600;
	time_t start = 0, now;
	int c;
#ifndef __APPLE__
	char details = 0;
#endif
	char f_nr = 0;

	fprintf(stderr, "acctail v1.3, (C) folkert@vanheusden.com\n\n");

#ifdef __APPLE__
	while((c = getopt(argc, argv, "s:h")) != -1)
#else
	while((c = getopt(argc, argv, "s:dh")) != -1)
#endif
	{
		switch(c)
		{
		case 's':
			delay = atoi(optarg);
			break;
#ifndef __APPLE__
		case 'd':
			details = 1;
			break;
#endif
		case 'h':
		default:
			usage();
			exit(1);
		}
	}

	signal(SIGTERM, sigh);
	signal(SIGINT, sigh);

#ifndef __APPLE__
	printf("COMMAND          FLGS UID     .GID      EXIT    UTIME   STIME   ETIME   MEM\n");
	if (details)
		printf("                 IO     RW     MINFLT MAJFLT SWAPS\n");
#else
	printf("COMMAND          FLGS UID     .GID      UTIME   STIME   ETIME   MEM   IO\n");
#endif
	printf("-------------------------------------------------------------------------------\n");

	for(;;)
	{
		struct acct record;
		int rc;
		char flags[5], nflags=0;
		struct passwd *pwent;
		struct group *grpent;

		time(&now);

		if ((now - start) > delay || fd == -1)
		{
			char *dummy = gen_temp_fname(f_nr);
			int fdd;

			f_nr = 1 - f_nr;

			fdd = open(dummy, O_RDWR | O_CREAT);
			if (fdd == -1)
			{
				fprintf(stderr, "cannot create file %s: %d\n", dummy, errno);
				exit(1);
			}

			if (acct(dummy) == -1)
			{
				fprintf(stderr, "cannot open %s: %d\n", dummy, errno);
				exit(1);
			}

			if (fname)
			{
				close(fd);
				if (unlink(fname) == -1)
				{
					fprintf(stderr, "error deleting file %s: %d\n", fname, errno);
					exit(1);
				}
				free(fname);
			}

			fname = dummy;
			fd = fdd;

			start = now;
		}

		rc = read(fd, &record, sizeof(record));

		if (rc == -1)
		{
			if (errno != EINTR)
			{
				fprintf(stderr, "read error (fd: %d): %d\n", fd, errno);
				exit(1);
			}
		}
		else if (rc == 0)
		{
			usleep(100);
			continue;
		}
		else if (rc != sizeof(record))
		{
			fprintf(stderr, "short read while retrieving record (got: %d, expected: %d)\n", rc, sizeof(record));
			exit(1);
		}

		if (record.ac_flag & AFORK)
			flags[nflags++] = 'F';
		if (record.ac_flag & ASU)
			flags[nflags++] = 'U';
		if (record.ac_flag & ACORE)
			flags[nflags++] = 'C';
		if (record.ac_flag & AXSIG)
			flags[nflags++] = 'S';
		flags[nflags] = 0x00;

		pwent = getpwuid(record.ac_uid);
		grpent = getgrgid(record.ac_gid);

#ifndef __APPLE__
		printf("%-16s %-4s %-8s.%-8s %-5d %7.2f %7.2f %7.2f %6ld\n",
#else
		printf("%-16s %-4s %-8s.%-8s %7.2f %7.2f %7.2f %6ld %6ld\n",
#endif
			record.ac_comm,
			flags,
			pwent -> pw_name,
			grpent -> gr_name,
#ifndef __APPLE__
			record.ac_exitcode,
#endif
			compt_to_msec(record.ac_utime),
			compt_to_msec(record.ac_stime),
			compt_to_msec(record.ac_etime),
			expand_compt(record.ac_mem)
#ifdef __APPLE__
			, expand_compt(record.ac_io)
#endif
		);
#ifndef __APPLE__
		if (details)
		{
			printf("                 %6ld %6ld %6ld %6ld %6ld\n",
				expand_compt(record.ac_io),
				expand_compt(record.ac_rw),
				expand_compt(record.ac_minflt),
				expand_compt(record.ac_majflt),
				expand_compt(record.ac_swaps)
			);
		}
#endif
	}

	return 0;
}
